@@ -1,9 +1,10 @@
 (* camlp5r *)
-(* $Id: pa_r.ml,v 6.36 2011-03-15 13:49:12 deraugla Exp $ *)
-(* Copyright (c) INRIA 2007-2011 *)
+(* pa_r.ml,v *)
+(* Copyright (c) INRIA 2007-2014 *)
 
 #load "pa_extend.cmo";
 #load "q_MLast.cmo";
+#load "pa_macro.cmo";
 
 open Pcaml;
 
@@ -12,9 +13,12 @@ Pcaml.no_constructors_arity.val := False;
 
 do {
   let odfa = Plexer.dollar_for_antiquotation.val in
+  let odni = Plexer.dot_newline_is.val in
   Plexer.dollar_for_antiquotation.val := False;
   Plexer.utf8_lexing.val := True;
+  Plexer.dot_newline_is.val := ";";
   Grammar.Unsafe.gram_reinit gram (Plexer.gmake ());
+  Plexer.dot_newline_is.val := odni;
   Plexer.dollar_for_antiquotation.val := odfa;
   Grammar.Unsafe.clear_entry interf;
   Grammar.Unsafe.clear_entry implem;
@@ -121,48 +125,6 @@ value rec generalized_type_of_type =
       ([], t) ]
 ;
 
-value start_with s s_ini =
-  let len = String.length s_ini in
-  String.length s >= len && String.sub s 0 len = s_ini
-;
-
-value greek_tab =
-  ["α"; "β"; "γ"; "δ"; "ε"; "ζ"; "η"; "θ"; "ι"; "κ"; "λ"; "μ"; "ν"; "ξ";
-   "ο"; "π"; "ρ"; "σ"; "τ"; "υ"; "φ"; "χ"; "ψ"; "ω"]
-;
-value index_tab = [""; "₁"; "₂"; "₃"; "₄"; "₅"; "₆"; "₇"; "₈"; "₉"];
-value ascii_of_greek s =
-  loop 0 greek_tab where rec loop i =
-    fun
-    [ [g :: gl] -> do {
-        if start_with s g then do {
-          let c1 = Char.chr (Char.code 'a' + i) in
-          let glen = String.length g in
-          let rest = String.sub s glen (String.length s - glen) in
-          loop 0 index_tab where rec loop i =
-            fun
-            [ [k :: kl] -> do {
-                if rest = k then do {
-                  let s2 = if i = 0 then "" else string_of_int i in
-                  String.make 1 c1 ^ s2
-                }
-                else loop (i + 1) kl
-              }
-            | [] -> s ]
-        }
-        else loop (i + 1) gl
-      }
-    | [] -> s ]
-;
-
-(* should be added in lib/plexer.ml, perhaps, as a new token GREEK? *)
-value greek_token =
-  Grammar.Entry.of_parser gram "greek_token"
-    (parser
-       [: `("LIDENT", x) when List.exists (start_with x) greek_tab :] ->
-          ascii_of_greek x)
-;
-
 value warned = ref False;
 value warning_deprecated_since_6_00 loc =
   if not warned.val then do {
@@ -172,6 +134,8 @@ value warning_deprecated_since_6_00 loc =
   else ()
 ;
 
+(* -- begin copy from pa_r to q_MLast -- *)
+
 EXTEND
   GLOBAL: sig_item str_item ctyp patt expr module_type module_expr
     signature structure class_type class_expr class_sig_item class_str_item
@@ -211,8 +175,8 @@ EXTEND
       | "include"; me = module_expr -> <:str_item< include $me$ >>
       | "module"; r = V (FLAG "rec"); l = V (LIST1 mod_binding SEP "and") ->
           <:str_item< module $_flag:r$ $_list:l$ >>
-      | "module"; "type"; i = V UIDENT "uid" ""; mt = mod_type_fun_binding ->
-          <:str_item< module type $_uid:i$ = $mt$ >>
+      | "module"; "type"; i = V ident ""; mt = mod_type_fun_binding ->
+          <:str_item< module type $_:i$ = $mt$ >>
       | "open"; i = V mod_ident "list" "" -> <:str_item< open $_:i$ >>
       | "type"; tdl = V (LIST1 type_decl SEP "and") ->
           <:str_item< type $_list:tdl$ >>
@@ -279,8 +243,8 @@ EXTEND
       | "module"; rf = V (FLAG "rec");
         l = V (LIST1 mod_decl_binding SEP "and") ->
           <:sig_item< module $_flag:rf$ $_list:l$ >>
-      | "module"; "type"; i = V UIDENT "uid" ""; "="; mt = module_type ->
-          <:sig_item< module type $_uid:i$ = $mt$ >>
+      | "module"; "type"; i = V ident ""; "="; mt = module_type ->
+          <:sig_item< module type $_:i$ = $mt$ >>
       | "open"; i = V mod_ident "list" "" -> <:sig_item< open $_:i$ >>
       | "type"; tdl = V (LIST1 type_decl SEP "and") ->
           <:sig_item< type $_list:tdl$ >>
@@ -325,6 +289,9 @@ EXTEND
       | "match"; e = SELF; "with"; "["; l = V (LIST0 match_case SEP "|");
         "]" ->
           <:expr< match $e$ with [ $_list:l$ ] >>
+      | "match"; e = SELF; "with"; "|"; l = V (LIST0 match_case SEP "|");
+        "end" ->
+          <:expr< match $e$ with [ $_list:l$ ] >>
       | "match"; e = SELF; "with"; p1 = ipatt; "->"; e1 = SELF ->
           <:expr< match $e$ with $p1$ -> $e1$ >>
       | "try"; e = SELF; "with"; "["; l = V (LIST0 match_case SEP "|"); "]" ->
@@ -404,6 +371,7 @@ EXTEND
       | s = V STRING -> <:expr< $_str:s$ >>
       | s = V CHAR -> <:expr< $_chr:s$ >>
       | i = V LIDENT -> <:expr< $_lid:i$ >>
+      | i = V GIDENT -> <:expr< $_lid:i$ >>
       | i = V UIDENT -> <:expr< $_uid:i$ >>
       | "["; "]" -> <:expr< [] >>
       | "["; el = LIST1 expr SEP ";"; last = cons_expr_opt; "]" ->
@@ -483,6 +451,7 @@ EXTEND
       [ p1 = SELF; "."; p2 = SELF -> <:patt< $p1$ . $p2$ >> ]
     | "simple"
       [ s = V LIDENT -> <:patt< $_lid:s$ >>
+      | s = V GIDENT -> <:patt< $_lid:s$ >>
       | s = V UIDENT -> <:patt< $_uid:s$ >>
       | s = V INT -> <:patt< $_int:s$ >>
       | s = V INT_l -> <:patt< $_int32:s$ >>
@@ -530,13 +499,15 @@ EXTEND
       [ p1 = SELF; "."; p2 = SELF -> <:patt< $p1$ . $p2$ >> ]
     | "simple" RIGHTA
       [ i = V UIDENT -> <:patt< $_uid:i$ >>
-      | i = V LIDENT -> <:patt< $_lid:i$ >> ] ]
+      | i = V LIDENT -> <:patt< $_lid:i$ >>
+      | "_" -> <:patt< _ >> ] ]
   ;
   ipatt:
     [ [ "{"; lpl = V (LIST1 label_ipatt SEP ";"); "}" ->
           <:patt< { $_list:lpl$ } >>
       | "("; p = paren_ipatt; ")" -> p
       | s = V LIDENT -> <:patt< $_lid:s$ >>
+      | s = V GIDENT -> <:patt< $_lid:s$ >>
       | "_" -> <:patt< _ >> ] ]
   ;
   paren_ipatt:
@@ -573,7 +544,7 @@ EXTEND
   ;
   simple_type_parameter:
     [ [ "'"; i = ident -> Some i
-      | i = greek_token -> Some i
+      | i = GIDENT -> Some (greek_ascii_equiv i)
       | "_" -> None ] ]
   ;
   ctyp:
@@ -597,7 +568,7 @@ EXTEND
       [ t1 = SELF; "."; t2 = SELF -> <:ctyp< $t1$ . $t2$ >> ]
     | "simple"
       [ "'"; i = V ident "" -> <:ctyp< '$_:i$ >>
-      | i = greek_token -> <:ctyp< '$i$ >>
+      | i = GIDENT -> <:ctyp< '$greek_ascii_equiv i$ >>
       | "_" -> <:ctyp< _ >>
       | i = V LIDENT -> <:ctyp< $_lid:i$ >>
       | i = V UIDENT -> <:ctyp< $_uid:i$ >>
@@ -946,8 +917,55 @@ EXTEND
     [ [ "to" -> True
       | "downto" -> False ] ]
   ;
+  (* -- cut 1 begin -- *)
+  expr: [[]];
 END;
 
+IFDEF JOCAML OR COMPATIBLE_WITH_OLD_OCAML THEN
+  EXTEND
+    GLOBAL: str_item expr;
+    (* -- cut 1 end -- *)
+    str_item: 
+      [ [ "def"; jal = V (LIST1 joinautomaton SEP "and") ->
+            <:str_item< def $_list:jal$ >> ] ]
+    ;
+    expr: LEVEL "top"
+      [ [ "def"; jal = V (LIST1 joinautomaton SEP "and"); "in";
+          e = expr LEVEL "top"->
+            <:expr< def $_list:jal$ in $e$ >> ] ]
+    ;
+    expr: LEVEL "apply"
+      [ [ "reply"; eo = V (OPT expr); "to"; ji = joinident ->
+            <:expr< reply $_opt:eo$ to $jid:ji$ >> ] ]
+    ;
+    expr: BEFORE ":="
+      [ [ "spawn"; e = SELF -> <:expr< spawn $e$ >> ] ]
+    ;
+    expr: LEVEL "&&"
+      [ [ e1 = SELF; "&"; e2 = SELF -> <:expr< $e1$ & $e2$ >> ] ]
+    ;
+    joinautomaton:
+      [ [ jcl = V (LIST1 joinclause SEP "or") ->
+            {MLast.jcLoc = loc; MLast.jcVal = jcl} ] ]
+    ;
+    joinclause:
+      [ [ jpl = V (LIST1 joinpattern SEP "&"); "="; e = expr ->
+            (loc, jpl, e) ] ]
+    ;
+    joinpattern:
+      [ [ ji = joinident; "("; op = V (OPT patt); ")" -> (loc, ji, op) ] ]
+    ;
+    joinident:
+      [ [ i = V LIDENT -> (loc, i) ] ]
+    ;
+    (* -- cut 2 begin -- *)
+    expr: [[]];
+  END;
+END;
+
+(* -- cut 2 end -- *)
+(* -- end copy from pa_r to q_MLast -- *)
+
 value quotation_content s = do {
   loop 0 where rec loop i =
     if i = String.length s then ("", s)
